class Operation
  module Deferrable
    attr_reader :arguments

    # Public Callbacks

    def on_start(&block)
      return unless block

      if started?
        block.call
      else
        startables.unshift block
      end
      self
    end

    def on_progress(&block)
      return unless block
      progressables.unshift block
      self
    end

    def on_success(&block)
      return unless block

      if succeeded?
        block.call(*arguments)
      elsif !failed?
        successables.unshift block
      end
      self
    end

    def on_failure(&block)
      return unless block

      if failed?
        block.call(*arguments)
      elsif !succeeded?
        failables.unshift block
      end
      self
    end

    def on_finish(&block)
      return unless block

      if succeeded? || failed?
        block.call
      else
        finishables.unshift block
      end
      self
    end

    # Public State Getters

    def status
      @status ||= :unknown
    end

    # E.g. 42%
    def percent
      return 100 if succeeded?
      @percent ||= 0
    end

    # E.g. 0.42
    def perone
      (percent.to_f / 100).round(2)
    end

    def method_name

    end

    def succeeded?
      status == :succeeded
    end

    def failed?
      status == :failed
    end

    def started?
      status != :unknown
    end

    # Public State Setters

    def progress!(new_percent, code = nil)
      return if succeeded? || failed?
      @percent = new_percent.to_i
      start! unless started?
      @status = code unless [:succeeded, :failed].include?(code)

      progressables.each do |callback|
        callback.call self
      end
      self
    end

    def sub_progress!(from, till, sub)
      if sub.percent.zero?
        new_percent = 0
      else
        delta = (till.to_f - from.to_f) * sub.perone
        new_percent = from.to_i + delta.to_i
      end
      progress! new_percent, sub.status
      self
    end

    def start!(*args)
      set_status :started, *args
    end

    def succeed!(*args)
      set_status :succeeded, *args
    end

    def fail!(*args)
      set_status :failed, *args
    end

    # Status Setters

    def set_status(new_status, *args)
      @arguments = args unless args.empty?
      @status = new_status

      call_startables!
      if succeeded? && !successables.empty?
        while callback = successables.pop
          callback.call(*arguments)
        end
        call_finishables!
        failables.clear if !failables.empty?

      elsif failed? && !failables.empty?
        while callback = failables.pop
          callback.call(*arguments)
        end
        call_finishables!
        successables.clear if !failables.empty?
      end
      self
    end

    def call_startables!
      return unless startables
      return if startables.empty?

      while callback = startables.pop
        callback.call
      end
    end

    def call_finishables!
      return unless finishables
      return if finishables.empty?

      while callback = finishables.pop
        callback.call
      end
    end

    # Internal Methods

    def startables
      @startables ||= []
    end

    def successables
      @successables ||= []
    end

    def failables
      @failables ||= []
    end

    def progressables
      @progressables ||= []
    end

    def finishables
      @finishables ||= []
    end

  end
end

class Operation
  class Defer
    include ::Operation::Deferrable
  end
end

